#ifndef STIM_GL_SPIDER_H
#define STIM_GL_SPIDER_H

#include <GL/glew.h>
#include <GL/glut.h>
#include <cuda.h>
#include <cuda_gl_interop.h>
#include <cudaGL.h>
#include <math.h>
#include "../gl/gl_texture.h"
#include "../visualization/camera.h"
#include "./error.h"
#include "../math/vector.h"
#include "../math/rect.h"
#include "../cuda/cost.h"
#include "../cuda/glbind.h"

#include <iostream>
#include <fstream>

namespace stim
{

template<typename T>
class gl_spider : public virtual gl_texture<T>
{
	//doen't use gl_texture really, just needs the GLuint id.
	//doesn't even need the texture iD really.
	private:
		stim::vec<float> position;  	//vector designating the position of the spider.
		stim::vec<float> direction;	//vector designating the orientation of the spider
						//always a unit vector.
		stim::vec<float> magnitude;	//magnitude of the direction vector.
						//mag[0] = length.
						//mag[1] = width.
		using gl_texture<T>::texID;
		//using image_stack<T>::S;
		cudaArray* c_Array;
		//void** devPtr;
		//size_t size;
		cudaGraphicsResource_t resource;
		GLuint fboID;
		GLuint texbufferID;
		int iter = 0;			//temporary for testing

		void
		findOptimalPosition()
		{
			/* Method for finding the best direction for the spider.
			   Not sure if necessary since the next position for the spider
			   will be at direction * magnitude. */
		}
	
		void
		findOptimalScale()
		{
			/* Method for finding the best scale for the spider.
			   changes the x, y, z size of the spider to minimize the cost
			   function. */
		}

		void
		Evaluate()
		{	
			/* Uses uniform sampler2D in order to take a difference between
			   the colors of two textures. 1st texture is the spider template,
			   the 2nd is the location of the spider's overlap with the
			   gl_template 
	
			   does the spider need to track it's location? Prob not since
			   position can be set with gl_texture coordinates */
		
		}
		
		void
		Optimize()
		{
			/*find the optimum direction and scale */ 
		}
		/*
		void
		Step()
		{
			// move to the new position 
		}
		*/
		
		void
		GenerateFBO(unsigned int width, unsigned int height)
		{
			glGenFramebuffers(1, &fboID);
			glBindFramebuffer(GL_FRAMEBUFFER, fboID);
			int numChannels = 1;
			unsigned char* texels = new unsigned char[width * height * numChannels];
			glGenTextures(1, &texbufferID);
			glBindTexture(GL_TEXTURE_2D, texbufferID);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE,
				 width, height, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, texels);   
			delete[] texels;
			glBindFramebuffer(GL_FRAMEBUFFER, 0); 
			glBindTexture(GL_TEXTURE_2D, 0);
		}

		void
		UpdateBuffer(float v_x, float v_y)
		{	
			//std::cout << v_y << std::endl;
			float len = 10.0;
			stim::vec<float>p1; 
        	        stim::vec<float>p2; 
	                stim::vec<float>p3; 
                	stim::vec<float>p4;	
			p1 = hor.p(1,1);
			p2 = hor.p(1,0);
			p3 = hor.p(0,0);
			p4 = hor.p(0,1);
			glBegin(GL_QUADS);
				glTexCoord3f(
					p1[0],
					p1[1],
					p1[2]
					);
				//glVertex2f(0.0,0.0);
				glVertex2f(v_x,v_y);
				glTexCoord3f(
					p2[0],
					p2[1],
					p2[2]
					);
				//glVertex2f(1.0, 0.0);
				glVertex2f(v_x+len, v_y);
				glTexCoord3f(
					p3[0],
					p3[1],
					p3[2]
					);
				//glVertex2f(1.0, 2.0);
				glVertex2f(v_x+len, v_y+len);
				glTexCoord3f(
					p4[0],
					p4[1],
					p4[2]
					);
				//glVertex2f(0.0, 2.0);
				glVertex2f(v_x, v_y+len);
			 glEnd();
			 p1 = ver.p(1,1);
			 p2 = ver.p(1,0);
			 p3 = ver.p(0,0);
			 p4 = ver.p(0,1);
		 	 glBegin(GL_QUADS);
				glTexCoord3f(
					p1[0],
					p1[1],
					p1[2]
					);
				//glVertex2f(1.0, 0.0);
				glVertex2f(v_x+len, v_y);
				glTexCoord3f(
					p2[0],
					p2[1],
					p2[2]
					);
				//glVertex2f(2.0, 0.0);
				glVertex2f(v_x+2*len, v_y);
				glTexCoord3f(
					p3[0],
					p3[1],
					p3[2]
					);
				//glVertex2f(2.0, 2.0);
				glVertex2f(v_x+2*len, v_y+len);
				glTexCoord3f(
					p4[0],
					p4[1],
					p4[2]
					);
				//glVertex2f(1.0, 2.0);
				glVertex2f(v_x+len, v_y+len);
			glEnd();
		}

		void
		UpdatePlanes(float v_x, float v_y, vec<float> vctr, int type = 0)
		{
			vec<float> Y(1.0,0.0,0.0);
			if(cos(Y.dot(vctr))< 0.087){                                                                             Y[0] = 0.0; Y[1] = 1.0;}
                        switch(type) {
				case 0:
                                         hor = stim::rect<float>(magnitude, position, vctr.norm(),
                        			((Y.cross(vctr)).cross(vctr)).norm());
              				 ver = stim::rect<float>(magnitude, position, vctr.norm(),
                        			hor.n());
               				 break;
        			case 1:
                			 hor = stim::rect<float>(magnitude, vctr, direction,
                      				((Y.cross(direction)).cross(direction)).norm());
                			 ver = stim::rect<float>(magnitude, vctr, direction,
                        			hor.n());
                			 break;
        			case 2:
                			 hor = stim::rect<float>(vctr, position, direction,
                      				((Y.cross(direction)).cross(direction)).norm());
                			 ver = stim::rect<float>(vctr, position, direction,
                        			hor.n());
					 break;
        			default:
                			 std::cout << "unknown case have been passed"
                         			<< std::endl;
					 break;
			}               
			UpdateBuffer(v_x, v_y);
		}
		
		/*
		Method for controling the buffer and texture binding in order to properly
		do the render to texture.
		*/
		void
		Bind(int numSamples = 1089)
		{
			float len = 10.0;
			glBindFramebuffer(GL_FRAMEBUFFER, fboID);//set up GL buffer		
			glFramebufferTexture2D(
				GL_FRAMEBUFFER,
				GL_COLOR_ATTACHMENT0,
				GL_TEXTURE_2D,
				texbufferID,
				0);
			glBindFramebuffer(GL_FRAMEBUFFER, fboID);
			GLenum DrawBuffers[1] = {GL_COLOR_ATTACHMENT0};
			glDrawBuffers(1, DrawBuffers);
			glBindTexture(GL_TEXTURE_2D, texbufferID);
			glClearColor(0,0,0,0);
			glClear(GL_COLOR_BUFFER_BIT);
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
			glViewport(0,0,2.0*len, numSamples*len);
			gluOrtho2D(0.0,2.0*len,0.0,numSamples*len);
			glEnable(GL_TEXTURE_3D);
			glBindTexture(GL_TEXTURE_3D, texID);
		}
		
		void
		Unbind()
		{
			//Finalize GL_buffer
			glBindTexture(GL_TEXTURE_3D, 0);                      
			glDisable(GL_TEXTURE_3D);
			glBindFramebuffer(GL_FRAMEBUFFER,0);
			glBindTexture(GL_TEXTURE_2D, 0);
		}
		/*
		Method for populating the buffer with the sampled texture.
		usually uses the default direction vector and then
		*/
		void
		sampleDirection(int numSamples = 1089, float solidAngle = M_PI)
		{
			
			//ofstream file;
			//file.open("samples.txt", std::ios_base::app);
			float samples[numSamples][3];    	//Set up the variables
							 	//necessary for sample generation
			vec<float> d_s = direction.cart2sph();
			vec<float> temp;
			int dim = (sqrt(numSamples)-1)/2;
			//std::cout << dim << std::endl;
			float y_0 	= 0.0;
			float p0  	= M_PI/3;
			float dt  	= solidAngle/(1.0 * dim);
			float dp  	= p0/(1.0*dim);
	
			Bind();	
			//file << "Step: Direction" << "\n";
			//file << "iteration: " << iter << "\n";
			//file << "starting pos and dir:" << "[" << position[0] << "," <<position[1] << "," << position[2] << "]" << ":" << "["  << direction[0] << "," << direction[1] << "," << direction[2] << "]\n"; 
			//Loop over the samples such that the original orientation is in the
			//center of the resulting texture index (544)
			//file << position[0] << "," <<position[1] << "," << position[2] << "\n"; 
			int idx;
			for(int i = -dim; i <= dim; i++){
				for(int j = -dim; j <= dim; j++){
					//Create linear index
					idx = (i+dim)*(dim*2+1) + (j+dim);

					temp[0] = d_s[0]; 			//rotate vector
					temp[1] = d_s[1]+dt*i;
					temp[2] = d_s[2]+dp*j;

					temp = (temp.sph2cart()).norm();		//back to cart
					samples[idx][0] = temp[0];	//save sample vector
					samples[idx][1] = temp[1];
					samples[idx][2] = temp[2];
	
			//		file << idx << ":"  <<"[" << samples[idx][0] << "," << samples[idx][1] << "," << samples[idx][2] << "]" << "\n";

					UpdatePlanes(0.0, y_0+(idx)*10, temp); 
				}
			}
			Unbind();
			int nxt = getCost();
			stim::vec<float> next(samples[nxt][0], samples[nxt][1], samples[nxt][2]);
			//next[0] = samples[nxt][0];
			//next[1] = samples[nxt][1];
			//next[2] = samples[nxt][2];
			next.norm();	
			//file << "next direction" << "[" << next[0] << "," << next[1] << "," << next[2] << "]\n\n"; 
			setPosition(position[0] + next[0]*magnitude[0]/2,
					 position[1]+next[1]*magnitude[0]/2,
						position[2]+next[2]*magnitude[0]/2);
			setDirection(next[0], next[1], next[2]);
			//file.close();
			setDirection(next);
			//file << "ending pos and dir:" << "[" << position[0] << "," <<position[1] << "," << position[2] << "]" << ":" << "["  << direction[0] << "," << direction[1] << "," << direction[2] << "]\n"; 
		}

		void
		samplePosition(int numSamples = 1089, float delta = 0.2)
		{
			
			float samples[numSamples][3];    	//Set up the variables
			vec<float> temp;
			int dim = (sqrt(numSamples)-1)/2;
			float y_0 	= 0.0;			//location of the first sample in the buffer
			stim::rect<float> samplingPlane =
				 stim::rect<float>(magnitude[0]*delta, position, direction);
			float step = 1.0/(dim);
			Bind();	
			//Loop over the samples, keeping the original position sample
			//in the center of the resulting texture.

			int idx;
			for(int i = -dim; i <= dim; i++){
				for(int j = -dim; j <= dim; j++){
					//Create linear index
					idx = (i+dim)*(dim*2+1) + (j+dim);

					temp = samplingPlane.p(
							0.5+step*i,
								 0.5+step*j
										);

					samples[idx][0] = temp[0];	//save sample vector
					samples[idx][1] = temp[1];
					samples[idx][2] = temp[2];
	
					UpdatePlanes(0.0, y_0+(idx)*10, temp, 1); 
				}
			}
			Unbind();
			int nxt = getCost();
			setPosition(samples[nxt][0], samples[nxt][1], samples[nxt][2]);
		}

		void
		sampleMagnitude(int numSamples = 1089, float delta = 0.5)
		{
			
			ofstream file;
			file.open("samples.txt", std::ios_base::app);
			float samples[numSamples];    	//Set up the variables
			int dim = (sqrt(numSamples)-1)/2;
			float y_0 	= 0.0;			//location of the first sample in the buffer
			float min 	= 1.0-delta;
			float max 	= 1.0+delta;
			float step	= (max-min)/(numSamples-1);
			float factor;
			vec<float> temp;
			Bind();	

			for(int i = 0; i < numSamples; i++){
					//Create linear index
					factor = (min+step*i)*magnitude[0];
					samples[i] = factor;	//save sample vector
					stim::vec<float> temp(factor);
					UpdatePlanes(0.0, y_0+(i)*10, temp, 2); 
			}
			Unbind();
			int nxt = getCost();
			setMagnitude(samples[nxt]);
			file << position[0] << "," <<position[1] << "," << position[2] << "\n"; 
		}

//--------------------------------------------------------------------------//
//--------------------------------CUDA METHODS------------------------------//
//--------------------------------------------------------------------------//
		
		/* Method for registering the texture with Cuda for shared
			access */
		void
		createResource()
		{
			HANDLE_ERROR(
				cudaGraphicsGLRegisterImage(
					 &resource,
				 	texbufferID,
				 	GL_TEXTURE_2D,
				 	//CU_GRAPHICS_REGISTER_FLAGS_NONE)
					cudaGraphicsMapFlagsReadOnly)
			);
		} 
		
		/* Method for freeing the texture from Cuda for gl access */
		void
		destroyResource()
		{
			HANDLE_ERROR(
				cudaGraphicsUnregisterResource(resource)
			);		
		}

		/* Entry-point into the cuda code for calculating the cost
			of a given samples array (in texture form) */
		int
		getCost()
		{
			createResource();
			int cost = 	get_cost(resource, iter);
			destroyResource();
			iter++;
			return cost;
		}

	public:

		stim::rect<float> hor;
		stim::rect<float> ver;	

		//Default Constructor
		gl_spider
		()
		{
			setPosition(0.0,0.0,0.0);
			setDirection(0.0,0.0,1.0);
			setMagnitude(0.03);
		}

		//temporary constructor for convenience, will be removed in further updates.	
		gl_spider
		(float pos_x, float pos_y, float pos_z, float dir_x, float dir_y, float dir_z,
			float mag_x)
		{
			setPosition(pos_x, pos_y, pos_z);
			setDirection(dir_x, dir_y, dir_z);
			setMagnitude(mag_x);
		}
		
		//Attached the spider to the texture with the given GLuint ID.
		//Samples in the default direction acting as the init method.	
		void
		attachSpider(GLuint id, int numSamples = 1089)
		{
			texID = id;
			GenerateFBO(20, numSamples*10);
			sampleDirection();
		}
		
		//temporary Method necessary for visualization and testing.
		void
		Update()
		{
			vec<float> Y(1.0,0.0,0.0);
			if(cos(Y.dot(direction))< 0.087){
				Y[0] = 0.0; Y[1] = 1.0;}
			hor = stim::rect<float>(magnitude, position, direction.norm(),
				((Y.cross(direction)).cross(direction)).norm());
			ver = stim::rect<float>(magnitude, position, direction.norm(),
				 hor.n());
		}

		//Returns the position vector.
		vec<float>
		getPosition()
		{
			return position;
		}
	
		//Returns the direction vector.
		vec<float>
		getDirection()
		{
			return direction;
		}

		//Returns the magnitude vector.
		vec<float>
		getMagnitude()
		{
			return magnitude;
		}
	
		//Sets the position vector to input vector pos
		void
		setPosition(vec<float> pos)
		{
			position = pos;
		}
		
		//Sets the position vector to the input float coordinates x,y,z
		void
		setPosition(float x, float y, float z)
		{
			position[0] = x;
			position[1] = y;
			position[2] = z;
		}

		
		//Sets the direction vector to input vector dir
		void
		setDirection(vec<float> dir)
		{
			direction = dir;
		}
		
		//Sets the direction vector to the input float coordinates x,y,z
		void
		setDirection(float x, float y, float z)
		{
			direction[0] = x;
			direction[1] = y;
			direction[2] = z;
		}
		
		//Sets the magnitude vector to the input vector mag.	
		void
		setMagnitude(vec<float> mag)
		{
			magnitude[0] = mag[0];
			magnitude[1] = mag[0];
		}
		
		void
		setMagnitude(float mag)
		{
			magnitude[0] = mag;
			magnitude[1] = mag;
		}
		
		//temporary method for visualization.
		GLuint
		getFB()
		{
			return fboID;
		}

		void
		Step()
		{
			findOptimalDirection();
		}


		void
		findOptimalDirection()
		{
			/* Method for finding the best direction for the spider.
			   Uses the camera to rotate. Then Calls Evaluate to find new cost.
			*/
			sampleDirection();
			samplePosition();
			sampleMagnitude();
		}

		/* Method for initializing the cuda devices, necessary only
			there are multiple cuda devices */
		void
		initCuda()
		{	
			//stim::cudaSetDevice();
			//GLint max;
			//glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max);
			//std::cout << max << std::endl;
		}
};
}
#endif
